@@ -26,13 +26,14 @@ import (
 	"os/exec"
 	"os/signal"
 	"path/filepath"
+	"strconv"
 	"syscall"
 	"time"
 
 	"github.com/spf13/pflag"
+	"k8s.io/apimachinery/pkg/types"
 	libvirt "libvirt.org/libvirt-go"
 
-	"k8s.io/apimachinery/pkg/types"
 	"k8s.io/apimachinery/pkg/util/wait"
 	utilwait "k8s.io/apimachinery/pkg/util/wait"
 	"k8s.io/apimachinery/pkg/watch"
@@ -122,7 +123,7 @@ func startDomainEventMonitoring(
 	virtShareDir string,
 	domainConn virtcli.Connection,
 	deleteNotificationSent chan watch.Event,
-	vmiUID types.UID,
+	vmi *v1.VirtualMachineInstance,
 	domainName string,
 	agentStore *agentpoller.AsyncAgentStore,
 	qemuAgentSysInterval time.Duration,
@@ -139,7 +140,7 @@ func startDomainEventMonitoring(
 		}
 	}()
 
-	err := notifier.StartDomainNotifier(domainConn, deleteNotificationSent, vmiUID, domainName, agentStore, qemuAgentSysInterval, qemuAgentFileInterval, qemuAgentUserInterval, qemuAgentVersionInterval)
+	err := notifier.StartDomainNotifier(domainConn, deleteNotificationSent, vmi, domainName, agentStore, qemuAgentSysInterval, qemuAgentFileInterval, qemuAgentUserInterval, qemuAgentVersionInterval)
 	if err != nil {
 		panic(err)
 	}
@@ -196,6 +197,11 @@ func initializeDirs(virtShareDir string,
 		panic(err)
 	}
 
+	err = virtlauncher.InitializeDisksDirectories(config.SysprepDisksDir)
+	if err != nil {
+		panic(err)
+	}
+
 	err = virtlauncher.InitializeDisksDirectories(config.SecretDisksDir)
 	if err != nil {
 		panic(err)
@@ -212,24 +218,40 @@ func initializeDirs(virtShareDir string,
 	}
 }
 
+func detectDomainWithUUID(domainManager virtwrap.DomainManager) *api.Domain {
+	domains, err := domainManager.ListAllDomains()
+	if err != nil {
+		log.Log.Reason(err).Errorf("failed to list domains when detecting UUID")
+		return nil
+	}
+	for _, domain := range domains {
+		if domain.Spec.UUID != "" {
+			return domain
+		}
+	}
+	return nil
+}
+
 func waitForDomainUUID(timeout time.Duration, events chan watch.Event, stop chan struct{}, domainManager virtwrap.DomainManager) *api.Domain {
 
 	ticker := time.NewTicker(timeout).C
 	checkEarlyExit := time.NewTicker(time.Second * 2).C
+	domainCheckTicker := time.NewTicker(time.Second * 10).C
 	for {
 		select {
 		case <-ticker:
 			panic(fmt.Errorf("timed out waiting for domain to be defined"))
-		case e := <-events:
-			if e.Object != nil && e.Type == watch.Added {
-				domain := e.Object.(*api.Domain)
-				log.Log.Infof("Detected domain with UUID %s", domain.Spec.UUID)
+		case <-domainCheckTicker:
+			log.Log.V(3).Infof("Periodically checking for domain with UUID")
+			domain := detectDomainWithUUID(domainManager)
+			if domain != nil {
+				return domain
+			}
+		case <-events:
+			log.Log.V(3).Infof("Checking for domain with UUID due to incoming libvirt event")
+			domain := detectDomainWithUUID(domainManager)
+			if domain != nil {
 				return domain
-			} else if e.Type == watch.Modified {
-				domain, ok := e.Object.(*api.Domain)
-				if ok && domain.ObjectMeta.DeletionTimestamp != nil {
-					return nil
-				}
 			}
 		case <-stop:
 			return nil
@@ -243,7 +265,7 @@ func waitForDomainUUID(timeout time.Duration, events chan watch.Event, stop chan
 
 func waitForFinalNotify(deleteNotificationSent chan watch.Event,
 	domainManager virtwrap.DomainManager,
-	vm *v1.VirtualMachineInstance) {
+	vmi *v1.VirtualMachineInstance) {
 
 	log.Log.Info("Waiting on final notifications to be sent to virt-handler.")
 
@@ -276,7 +298,7 @@ func waitForFinalNotify(deleteNotificationSent chan watch.Event,
 	// KillVMI is idempotent. Making a call to KillVMI here ensures that the deletion
 	// occurs regardless if the VirtualMachineInstance crashed unexpectedly or if virt-handler requested
 	// a graceful shutdown.
-	domainManager.KillVMI(vm)
+	domainManager.KillVMI(vmi)
 
 	// We don't want to block here forever. If the delete does not occur, that could mean
 	// something is wrong with libvirt. In this situation, virt-handler will detect that
@@ -317,7 +339,7 @@ func main() {
 	name := pflag.String("name", "", "Name of the VirtualMachineInstance")
 	uid := pflag.String("uid", "", "UID of the VirtualMachineInstance")
 	namespace := pflag.String("namespace", "", "Namespace of the VirtualMachineInstance")
-	gracePeriodSeconds := pflag.Int("grace-period-seconds", 30, "Grace period to observe before sending SIGTERM to vm process")
+	gracePeriodSeconds := pflag.Int("grace-period-seconds", 30, "Grace period to observe before sending SIGTERM to vmi process")
 	useEmulation := pflag.Bool("use-emulation", false, "Use software emulation")
 	hookSidecars := pflag.Uint("hook-sidecars", 0, "Number of requested hook sidecars, virt-launcher will wait for all of them to become available")
 	noFork := pflag.Bool("no-fork", false, "Fork and let virt-launcher watch itself to react to crashes if set to false")
@@ -335,6 +357,16 @@ func main() {
 
 	log.InitializeLogging("virt-launcher")
 
+	// check if virt-launcher verbosity should be changed
+	if verbosityStr, ok := os.LookupEnv("VIRT_LAUNCHER_LOG_VERBOSITY"); ok {
+		if verbosity, err := strconv.Atoi(verbosityStr); err == nil {
+			log.Log.SetVerbosityLevel(verbosity)
+			log.Log.V(2).Infof("set log verbosity to %d", verbosity)
+		} else {
+			log.Log.Warningf("failed to set log verbosity. The value of logVerbosity label should be an integer, got %s instead.", verbosityStr)
+		}
+	}
+
 	if !*noFork {
 		exitCode, err := ForkAndMonitor(*containerDiskDir)
 		if err != nil {
@@ -351,7 +383,7 @@ func main() {
 		panic(err)
 	}
 
-	vm := v1.NewVMIReferenceFromNameWithNS(*namespace, *name)
+	vmi := v1.NewVMIReferenceWithUUID(*namespace, *name, types.UID(*uid))
 
 	// Initialize local and shared directories
 	initializeDirs(*virtShareDir, *ephemeralDiskDir, *containerDiskDir, *hotplugDiskDir, *uid)
@@ -365,7 +397,7 @@ func main() {
 	}
 	util.StartLibvirt(stopChan)
 	// only single domain should be present
-	domainName := api.VMINamespaceKeyFunc(vm)
+	domainName := api.VMINamespaceKeyFunc(vmi)
 	util.StartVirtlog(stopChan, domainName)
 
 	domainConn := createLibvirtConnection()
@@ -390,7 +422,7 @@ func main() {
 
 	gracefulShutdownCallback := func() {
 		err := wait.PollImmediate(time.Second, 15*time.Second, func() (bool, error) {
-			err := domainManager.MarkGracefulShutdownVMI(vm)
+			err := domainManager.MarkGracefulShutdownVMI(vmi)
 			if err != nil {
 				log.Log.Reason(err).Errorf("Unable to signal graceful shutdown")
 				return false, err
@@ -403,11 +435,11 @@ func main() {
 			log.Log.Reason(err).Errorf("Gave up attempting to signal graceful shutdown")
 		}
 
-		log.Log.Object(vm).Info("Successfully signaled graceful shutdown")
+		log.Log.Object(vmi).Info("Successfully signaled graceful shutdown")
 	}
 
 	finalShutdownCallback := func(pid int) {
-		err := domainManager.KillVMI(vm)
+		err := domainManager.KillVMI(vmi)
 		if err != nil {
 			log.Log.Reason(err).Errorf("Unable to stop qemu with libvirt, falling back to SIGTERM")
 			syscall.Kill(pid, syscall.SIGTERM)
@@ -416,7 +448,7 @@ func main() {
 
 	events := make(chan watch.Event, 2)
 	// Send domain notifications to virt-handler
-	startDomainEventMonitoring(notifier, *virtShareDir, domainConn, events, vm.UID, domainName, &agentStore, *qemuAgentSysInterval, *qemuAgentFileInterval, *qemuAgentUserInterval, *qemuAgentVersionInterval)
+	startDomainEventMonitoring(notifier, *virtShareDir, domainConn, events, vmi, domainName, &agentStore, *qemuAgentSysInterval, *qemuAgentFileInterval, *qemuAgentUserInterval, *qemuAgentVersionInterval)
 
 	c := make(chan os.Signal, 1)
 	signal.Notify(c, os.Interrupt,
@@ -452,7 +484,7 @@ func main() {
 		// Now that the pid has exited, we wait for the final delete notification to be
 		// sent back to virt-handler. This delete notification contains the reason the
 		// domain exited.
-		waitForFinalNotify(events, domainManager, vm)
+		waitForFinalNotify(events, domainManager, vmi)
 	}
 
 	close(stopChan)
@@ -474,7 +506,7 @@ func ForkAndMonitor(containerDiskDir string) (int, error) {
 		return 1, err
 	}
 
-	exitStatus := make(chan syscall.WaitStatus, 10)
+	exitStatus := make(chan int, 10)
 	sigs := make(chan os.Signal, 10)
 	signal.Notify(sigs, syscall.SIGINT, syscall.SIGTERM, syscall.SIGQUIT, syscall.SIGCHLD)
 	go func() {
@@ -487,12 +519,8 @@ func ForkAndMonitor(containerDiskDir string) (int, error) {
 					log.Log.Reason(err).Errorf("Failed to reap process %d", wpid)
 				}
 
-				// there's a race between cmd.Wait() and syscall.Wait4 when
-				// cleaning up the cmd's pid after it exits. This allows us
-				// to detect the correct exit code regardless of which wait
-				// wins the race.
 				if wpid == cmd.Process.Pid {
-					exitStatus <- wstatus
+					exitStatus <- wstatus.ExitStatus()
 				}
 
 			default:
@@ -506,22 +534,9 @@ func ForkAndMonitor(containerDiskDir string) (int, error) {
 		}
 	}()
 
-	// wait for virt-launcher and collect the exit code
-	exitCode := 0
-	if err := cmd.Wait(); err != nil {
-		select {
-		case status := <-exitStatus:
-			exitCode = int(status)
-		default:
-			exitCode = 1
-			if exiterr, ok := err.(*exec.ExitError); ok {
-				if status, ok := exiterr.Sys().(syscall.WaitStatus); ok {
-					exitCode = status.ExitStatus()
-				}
-			}
-			log.Log.Reason(err).Error("dirty virt-launcher shutdown")
-		}
-
+	exitCode := <-exitStatus
+	if exitCode != 0 {
+		log.Log.Errorf("dirty virt-launcher shutdown: exit-code %d", exitCode)
 	}
 	// give qemu some time to shut down in case it survived virt-handler
 	// Most of the time we call `qemu-system=* binaries, but qemu-system-* packages
@@ -551,6 +566,9 @@ func ForkAndMonitor(containerDiskDir string) (int, error) {
 			}
 			return false, nil
 		})
+		if err != nil {
+			return 1, err
+		}
 	}
 	return exitCode, nil
 }
