@@ -20,6 +20,7 @@
 package watch
 
 import (
+	"context"
 	"encoding/json"
 	"fmt"
 	"reflect"
@@ -27,6 +28,7 @@ import (
 	"time"
 
 	k8sv1 "k8s.io/api/core/v1"
+	"k8s.io/api/policy/v1beta1"
 	v1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/labels"
 	"k8s.io/apimachinery/pkg/types"
@@ -35,6 +37,7 @@ import (
 	"k8s.io/client-go/tools/record"
 	"k8s.io/client-go/util/workqueue"
 
+	"kubevirt.io/kubevirt/pkg/util/pdbs"
 	"kubevirt.io/kubevirt/pkg/util/status"
 
 	virtconfig "kubevirt.io/kubevirt/pkg/virt-config"
@@ -48,6 +51,12 @@ import (
 	"kubevirt.io/kubevirt/pkg/virt-controller/services"
 )
 
+const (
+	failedToProcessDeleteNotificationErrMsg   = "Failed to process delete notification"
+	successfulUpdatePodDisruptionBudgetReason = "SuccessfulUpdate"
+	failedUpdatePodDisruptionBudgetReason     = "FailedUpdate"
+)
+
 type MigrationController struct {
 	templateService    services.TemplateService
 	clientset          kubecli.KubevirtClient
@@ -55,6 +64,7 @@ type MigrationController struct {
 	vmiInformer        cache.SharedIndexInformer
 	podInformer        cache.SharedIndexInformer
 	migrationInformer  cache.SharedIndexInformer
+	pdbInformer        cache.SharedIndexInformer
 	recorder           record.EventRecorder
 	podExpectations    *controller.UIDTrackingControllerExpectations
 	migrationStartLock *sync.Mutex
@@ -66,6 +76,7 @@ func NewMigrationController(templateService services.TemplateService,
 	vmiInformer cache.SharedIndexInformer,
 	podInformer cache.SharedIndexInformer,
 	migrationInformer cache.SharedIndexInformer,
+	pdbInformer cache.SharedIndexInformer,
 	recorder record.EventRecorder,
 	clientset kubecli.KubevirtClient,
 	clusterConfig *virtconfig.ClusterConfig,
@@ -77,6 +88,7 @@ func NewMigrationController(templateService services.TemplateService,
 		vmiInformer:        vmiInformer,
 		podInformer:        podInformer,
 		migrationInformer:  migrationInformer,
+		pdbInformer:        pdbInformer,
 		recorder:           recorder,
 		clientset:          clientset,
 		podExpectations:    controller.NewUIDTrackingControllerExpectations(controller.NewControllerExpectations()),
@@ -103,6 +115,10 @@ func NewMigrationController(templateService services.TemplateService,
 		UpdateFunc: c.updateMigration,
 	})
 
+	c.pdbInformer.AddEventHandler(cache.ResourceEventHandlerFuncs{
+		UpdateFunc: c.updatePDB,
+	})
+
 	return c
 }
 
@@ -112,7 +128,7 @@ func (c *MigrationController) Run(threadiness int, stopCh <-chan struct{}) {
 	log.Log.Info("Starting migration controller.")
 
 	// Wait for cache sync before we start the pod controller
-	cache.WaitForCacheSync(stopCh, c.vmiInformer.HasSynced, c.podInformer.HasSynced, c.migrationInformer.HasSynced)
+	cache.WaitForCacheSync(stopCh, c.vmiInformer.HasSynced, c.podInformer.HasSynced, c.migrationInformer.HasSynced, c.pdbInformer.HasSynced)
 
 	// Start the actual work
 	for i := 0; i < threadiness; i++ {
@@ -188,15 +204,14 @@ func (c *MigrationController) execute(key string) error {
 	}
 
 	if !vmiExists {
+		var err error
+
 		if migration.DeletionTimestamp == nil {
 			logger.V(3).Infof("Deleting migration for deleted vmi %s/%s", migration.Namespace, migration.Spec.VMIName)
-			err := c.clientset.VirtualMachineInstanceMigration(migration.Namespace).Delete(migration.Name, &v1.DeleteOptions{})
-			if err != nil {
-				return err
-			}
+			err = c.clientset.VirtualMachineInstanceMigration(migration.Namespace).Delete(migration.Name, &v1.DeleteOptions{})
 		}
 		// nothing to process for a migration that's being deleted
-		return nil
+		return err
 	}
 
 	vmi = vmiObj.(*virtv1.VirtualMachineInstance)
@@ -215,7 +230,7 @@ func (c *MigrationController) execute(key string) error {
 		syncErr = c.sync(key, migration, vmi, targetPods)
 	}
 
-	err = c.updateStatus(migration, vmi, targetPods, syncErr)
+	err = c.updateStatus(migration, vmi, targetPods)
 	if err != nil {
 		return err
 	}
@@ -262,7 +277,7 @@ func (c *MigrationController) canMigrateVMI(migration *virtv1.VirtualMachineInst
 
 }
 
-func (c *MigrationController) updateStatus(migration *virtv1.VirtualMachineInstanceMigration, vmi *virtv1.VirtualMachineInstance, pods []*k8sv1.Pod, syncErr error) error {
+func (c *MigrationController) updateStatus(migration *virtv1.VirtualMachineInstanceMigration, vmi *virtv1.VirtualMachineInstance, pods []*k8sv1.Pod) error {
 
 	var pod *k8sv1.Pod = nil
 	conditionManager := controller.NewVirtualMachineInstanceMigrationConditionManager()
@@ -315,7 +330,7 @@ func (c *MigrationController) updateStatus(migration *virtv1.VirtualMachineInsta
 
 		migrationCopy.Status.Phase = virtv1.MigrationFailed
 		c.recorder.Eventf(migration, k8sv1.EventTypeWarning, FailedMigrationReason, "Source node reported migration failed")
-		log.Log.Object(migration).Error("VMI reported migration failed.")
+		log.Log.Object(migration).Errorf("VMI %s/%s reported migration failed.", vmi.Namespace, vmi.Name)
 	} else if migration.DeletionTimestamp != nil && !migration.IsFinal() &&
 		!conditionManager.HasCondition(migration, virtv1.VirtualMachineInstanceMigrationAbortRequested) {
 		condition := virtv1.VirtualMachineInstanceMigrationCondition{
@@ -423,7 +438,7 @@ func (c *MigrationController) createTargetPod(migration *virtv1.VirtualMachineIn
 
 	key := controller.MigrationKey(migration)
 	c.podExpectations.ExpectCreations(key, 1)
-	pod, err := c.clientset.CoreV1().Pods(vmi.GetNamespace()).Create(templatePod)
+	pod, err := c.clientset.CoreV1().Pods(vmi.GetNamespace()).Create(context.Background(), templatePod, v1.CreateOptions{})
 	if err != nil {
 		c.recorder.Eventf(vmi, k8sv1.EventTypeWarning, FailedCreatePodReason, "Error creating pod: %v", err)
 		c.podExpectations.CreationObserved(key)
@@ -433,6 +448,41 @@ func (c *MigrationController) createTargetPod(migration *virtv1.VirtualMachineIn
 	return nil
 }
 
+func (c *MigrationController) expandPDB(pdb *v1beta1.PodDisruptionBudget, vmi *virtv1.VirtualMachineInstance, vmim *virtv1.VirtualMachineInstanceMigration) error {
+	minAvailable := 2
+
+	if pdb.Spec.MinAvailable.IntValue() == minAvailable && pdb.Labels[virtv1.MigrationNameLabel] == vmim.Name {
+		log.Log.V(4).Object(vmi).Infof("PDB has been already expanded")
+		return nil
+	}
+
+	patch := []byte(fmt.Sprintf(`{"spec":{"minAvailable": %d},"metadata":{"labels":{"%s": "%s"}}}`, minAvailable, virtv1.MigrationNameLabel, vmim.Name))
+
+	_, err := c.clientset.PolicyV1beta1().PodDisruptionBudgets(pdb.Namespace).Patch(context.Background(), pdb.Name, types.StrategicMergePatchType, patch, v1.PatchOptions{})
+	if err != nil {
+		c.recorder.Eventf(vmi, k8sv1.EventTypeWarning, failedUpdatePodDisruptionBudgetReason, "Error expanding the PodDisruptionBudget %s: %v", pdb.Name, err)
+		return err
+	}
+	log.Log.Object(vmi).Infof("expanding pdb for VMI %s/%s to protect migration %s", vmi.Namespace, vmi.Name, vmim.Name)
+	c.recorder.Eventf(vmi, k8sv1.EventTypeNormal, successfulUpdatePodDisruptionBudgetReason, "Expanded PodDisruptionBudget %s", pdb.Name)
+	return nil
+}
+
+func isMigrationProtected(pdb *v1beta1.PodDisruptionBudget) bool {
+	return pdb.Status.DesiredHealthy == 2 && pdb.Generation == pdb.Status.ObservedGeneration
+}
+
+func filterOutOldPDBs(pdbList []*v1beta1.PodDisruptionBudget) []*v1beta1.PodDisruptionBudget {
+	var filteredPdbs []*v1beta1.PodDisruptionBudget
+
+	for i := range pdbList {
+		if !pdbs.IsPDBFromOldMigrationController(pdbList[i]) {
+			filteredPdbs = append(filteredPdbs, pdbList[i])
+		}
+	}
+	return filteredPdbs
+}
+
 func (c *MigrationController) sync(key string, migration *virtv1.VirtualMachineInstanceMigration, vmi *virtv1.VirtualMachineInstance, pods []*k8sv1.Pod) error {
 
 	var pod *k8sv1.Pod = nil
@@ -441,14 +491,10 @@ func (c *MigrationController) sync(key string, migration *virtv1.VirtualMachineI
 		pod = pods[0]
 	}
 
-	if vmi == nil || vmi.DeletionTimestamp != nil {
-		// nothing to do with a deleted vmi
-		return nil
-	} else if vmi.Status.MigrationState != nil &&
-		vmi.Status.MigrationState.MigrationUID == migration.UID &&
-		vmi.Status.MigrationState.EndTimestamp != nil {
+	vmiDeleted := vmi == nil || vmi.DeletionTimestamp != nil
+	migrationFinalizedOnVMI := vmi.Status.MigrationState != nil && vmi.Status.MigrationState.MigrationUID == migration.UID && vmi.Status.MigrationState.EndTimestamp != nil
 
-		// nothing to do here, the migration is done
+	if vmiDeleted || migrationFinalizedOnVMI {
 		return nil
 	}
 
@@ -511,6 +557,31 @@ func (c *MigrationController) sync(key string, migration *virtv1.VirtualMachineI
 			// migration was accepted into the system, now see if we
 			// should create the target pod
 			if vmi.IsRunning() {
+				if migrations.MigrationNeedsProtection(vmi) {
+					pdbs, err := pdbs.PDBsForVMI(vmi, c.pdbInformer)
+					if err != nil {
+						return err
+					}
+					// removes pdbs from old implementation from list.
+					pdbs = filterOutOldPDBs(pdbs)
+
+					if len(pdbs) < 1 {
+						log.Log.Object(vmi).Errorf("Found no PDB protecting the vmi")
+						return fmt.Errorf("Found no PDB protecting the vmi %s", vmi.Name)
+					}
+					pdb := pdbs[0]
+
+					if err := c.expandPDB(pdb, vmi, migration); err != nil {
+						return err
+					}
+
+					// before proceeding we have to check that the k8s pdb controller has processed
+					// the pdb expansion and is actually protecting the VMI migration
+					if !isMigrationProtected(pdb) {
+						log.Log.V(4).Object(migration).Infof("Waiting for the pdb-controller to protect the migration pods, postponing migration start")
+						return nil
+					}
+				}
 				return c.createTargetPod(migration, vmi)
 			}
 			return nil
@@ -548,6 +619,7 @@ func (c *MigrationController) sync(key string, migration *virtv1.VirtualMachineI
 		}
 	case virtv1.MigrationPreparingTarget, virtv1.MigrationTargetReady, virtv1.MigrationFailed:
 		if (!podExists || podIsDown(pod)) &&
+			vmi.Status.MigrationState != nil &&
 			len(vmi.Status.MigrationState.TargetDirectMigrationNodePorts) == 0 &&
 			vmi.Status.MigrationState.StartTimestamp == nil &&
 			!vmi.Status.MigrationState.Failed &&
@@ -621,6 +693,9 @@ func (c *MigrationController) listMatchingTargetPods(migration *virtv1.VirtualMa
 			virtv1.MigrationJobLabel: string(migration.UID),
 		},
 	})
+	if err != nil {
+		return nil, err
+	}
 
 	objs, err := c.podInformer.GetIndexer().ByIndex(cache.NamespaceIndex, migration.Namespace)
 	if err != nil {
@@ -646,7 +721,7 @@ func (c *MigrationController) deleteMigration(obj interface{}) {
 	c.enqueueMigration(obj)
 }
 
-func (c *MigrationController) updateMigration(old, curr interface{}) {
+func (c *MigrationController) updateMigration(_, curr interface{}) {
 	c.enqueueMigration(curr)
 }
 
@@ -775,12 +850,12 @@ func (c *MigrationController) deletePod(obj interface{}) {
 	if !ok {
 		tombstone, ok := obj.(cache.DeletedFinalStateUnknown)
 		if !ok {
-			log.Log.Reason(fmt.Errorf("couldn't get object from tombstone %+v", obj)).Error("Failed to process delete notification")
+			log.Log.Reason(fmt.Errorf("couldn't get object from tombstone %+v", obj)).Error(failedToProcessDeleteNotificationErrMsg)
 			return
 		}
 		pod, ok = tombstone.Obj.(*k8sv1.Pod)
 		if !ok {
-			log.Log.Reason(fmt.Errorf("tombstone contained object that is not a pod %#v", obj)).Error("Failed to process delete notification")
+			log.Log.Reason(fmt.Errorf("tombstone contained object that is not a pod %#v", obj)).Error(failedToProcessDeleteNotificationErrMsg)
 			return
 		}
 	}
@@ -798,6 +873,34 @@ func (c *MigrationController) deletePod(obj interface{}) {
 	c.enqueueMigration(migration)
 }
 
+func (c *MigrationController) updatePDB(old, cur interface{}) {
+	curPDB := cur.(*v1beta1.PodDisruptionBudget)
+	oldPDB := old.(*v1beta1.PodDisruptionBudget)
+	if curPDB.ResourceVersion == oldPDB.ResourceVersion {
+		return
+	}
+
+	// Only process PDBs manipulated by this controller
+	migrationName := curPDB.Labels[virtv1.MigrationNameLabel]
+	if migrationName == "" {
+		return
+	}
+
+	objs, err := c.migrationInformer.GetIndexer().ByIndex(cache.NamespaceIndex, curPDB.Namespace)
+	if err != nil {
+		return
+	}
+
+	for _, obj := range objs {
+		vmim := obj.(*virtv1.VirtualMachineInstanceMigration)
+
+		if vmim.Name == migrationName {
+			log.Log.V(4).Object(curPDB).Infof("PDB updated")
+			c.enqueueMigration(vmim)
+		}
+	}
+}
+
 // takes a namespace and returns all migrations listening for this vmi
 func (c *MigrationController) listMigrationsMatchingVMI(namespace string, name string) ([]*virtv1.VirtualMachineInstanceMigration, error) {
 	objs, err := c.migrationInformer.GetIndexer().ByIndex(cache.NamespaceIndex, namespace)
@@ -872,12 +975,12 @@ func (c *MigrationController) deleteVMI(obj interface{}) {
 	if !ok {
 		tombstone, ok := obj.(cache.DeletedFinalStateUnknown)
 		if !ok {
-			log.Log.Reason(fmt.Errorf("couldn't get object from tombstone %+v", obj)).Error("Failed to process delete notification")
+			log.Log.Reason(fmt.Errorf("couldn't get object from tombstone %+v", obj)).Error(failedToProcessDeleteNotificationErrMsg)
 			return
 		}
 		vmi, ok = tombstone.Obj.(*virtv1.VirtualMachineInstance)
 		if !ok {
-			log.Log.Reason(fmt.Errorf("tombstone contained object that is not a vmi %#v", obj)).Error("Failed to process delete notification")
+			log.Log.Reason(fmt.Errorf("tombstone contained object that is not a vmi %#v", obj)).Error(failedToProcessDeleteNotificationErrMsg)
 			return
 		}
 	}
@@ -909,31 +1012,26 @@ func (c *MigrationController) outboundMigrationsOnNode(node string, runningMigra
 func (c *MigrationController) findRunningMigrations() ([]*virtv1.VirtualMachineInstanceMigration, error) {
 
 	// Don't start new migrations if we wait for migration object updates because of new target pods
-	notFinishedMigrations, err := migrations.ListUnfinishedMigrations(c.migrationInformer)
-	if err != nil {
-		return nil, err
-	}
-
+	notFinishedMigrations := migrations.ListUnfinishedMigrations(c.migrationInformer)
 	var runningMigrations []*virtv1.VirtualMachineInstanceMigration
 	for _, migration := range notFinishedMigrations {
 		if migration.IsRunning() {
 			runningMigrations = append(runningMigrations, migration)
-		} else {
-
-			vmi, exists, err := c.vmiInformer.GetStore().GetByKey(migration.Namespace + "/" + migration.Spec.VMIName)
-			if err != nil {
-				return nil, err
-			}
-			if !exists {
-				continue
-			}
-			pods, err := c.listMatchingTargetPods(migration, vmi.(*virtv1.VirtualMachineInstance))
-			if err != nil {
-				return nil, err
-			}
-			if len(pods) > 0 {
-				runningMigrations = append(runningMigrations, migration)
-			}
+			continue
+		}
+		vmi, exists, err := c.vmiInformer.GetStore().GetByKey(migration.Namespace + "/" + migration.Spec.VMIName)
+		if err != nil {
+			return nil, err
+		}
+		if !exists {
+			continue
+		}
+		pods, err := c.listMatchingTargetPods(migration, vmi.(*virtv1.VirtualMachineInstance))
+		if err != nil {
+			return nil, err
+		}
+		if len(pods) > 0 {
+			runningMigrations = append(runningMigrations, migration)
 		}
 	}
 	return runningMigrations, nil
