@@ -880,6 +880,30 @@ var _ = Describe("Manager", func() {
 
 			manager.MarkGracefulShutdownVMI(vmi)
 		})
+
+		It("Should signal graceful shutdown after marked for shutdown", func() {
+			mockDomain.EXPECT().Free().AnyTimes()
+
+			vmi := newVMI(testNamespace, testVmName)
+			domainSpec := expectIsolationDetectionForVMI(vmi)
+
+			xml, err := xml.MarshalIndent(domainSpec, "", "\t")
+			Expect(err).To(BeNil())
+
+			Expect(err).To(BeNil())
+			mockDomain.EXPECT().GetState().AnyTimes().Return(libvirt.DOMAIN_RUNNING, 1, nil)
+			mockConn.EXPECT().LookupDomainByName(testDomainName).AnyTimes().Return(mockDomain, nil)
+			mockDomain.EXPECT().GetXMLDesc(gomock.Eq(libvirt.DomainXMLFlags(0))).AnyTimes().Return(string(xml), nil)
+			mockDomain.EXPECT().
+				GetMetadata(libvirt.DOMAIN_METADATA_ELEMENT, "http://kubevirt.io", libvirt.DOMAIN_AFFECT_CONFIG).
+				AnyTimes().
+				Return(`<kubevirt><graceperiod><deletionGracePeriodSeconds>3600</deletionGracePeriodSeconds><deletionTimestamp>2021-03-11T09:08:20.144606353Z</deletionTimestamp><markedForGracefulShutdown>true</markedForGracefulShutdown></graceperiod></kubevirt>`, nil)
+
+			mockDomain.EXPECT().ShutdownFlags(libvirt.DOMAIN_SHUTDOWN_ACPI_POWER_BTN).Times(1).Return(nil)
+			manager, _ := NewLibvirtDomainManager(mockConn, testVirtShareDir, nil, 0, nil, "/usr/share/OVMF")
+
+			manager.SignalShutdownVMI(vmi)
+		})
 	})
 	Context("test migration monitor", func() {
 		It("migration should be canceled if it's not progressing", func() {
@@ -1710,6 +1734,79 @@ var _ = Describe("Manager", func() {
 			})
 		})
 	})
+	It("executes generateCloudInitEmptyISO and succeeds", func() {
+		agentStore := agentpoller.NewAsyncAgentStore()
+		agentStore.Store(agentpoller.GET_FILESYSTEM, []api.Filesystem{
+			{
+				Name:       "test",
+				Mountpoint: "/mnt/whatever",
+				Type:       "fs",
+				UsedBytes:  0,
+				TotalBytes: 0,
+			},
+		})
+
+		manager, _ := NewLibvirtDomainManager(mockConn, testVirtShareDir, nil, 0, &agentStore, "/usr/share/OVMF")
+
+		// we need the non-typecast object to make the function we want to test available
+		libvirtmanager := manager.(*LibvirtDomainManager)
+
+		vmi := newVMI(testNamespace, testVmName)
+		vmi.Status.VolumeStatus = make([]v1.VolumeStatus, 1)
+		vmi.Status.VolumeStatus[0] = v1.VolumeStatus{
+			Name: "test1",
+			Size: 42,
+		}
+
+		userData := "fake\nuser\ndata\n"
+		networkData := "FakeNetwork"
+		addCloudInitDisk(vmi, userData, networkData)
+		libvirtmanager.cloudInitDataStore = &cloudinit.CloudInitData{
+			DataSource: cloudinit.DataSourceNoCloud,
+			VolumeName: "test1",
+		}
+
+		err := libvirtmanager.generateCloudInitEmptyISO(vmi, nil)
+		Expect(err).ToNot(HaveOccurred())
+
+		isoPath := cloudinit.GetIsoFilePath(libvirtmanager.cloudInitDataStore.DataSource, vmi.Name, vmi.Namespace)
+		stats, err := os.Stat(isoPath)
+		Expect(err).ToNot(HaveOccurred())
+		Expect(stats.Size()).To(Equal(int64(42)))
+	})
+
+	It("executes generateCloudInitEmptyISO and fails", func() {
+		agentStore := agentpoller.NewAsyncAgentStore()
+		agentStore.Store(agentpoller.GET_FILESYSTEM, []api.Filesystem{
+			{
+				Name:       "test",
+				Mountpoint: "/mnt/whatever",
+				Type:       "fs",
+				UsedBytes:  0,
+				TotalBytes: 0,
+			},
+		})
+
+		manager, _ := NewLibvirtDomainManager(mockConn, testVirtShareDir, nil, 0, &agentStore, "/usr/share/OVMF")
+
+		// we need the non-typecast object to make the function we want to test available
+		libvirtmanager := manager.(*LibvirtDomainManager)
+
+		vmi := newVMI(testNamespace, testVmName)
+		vmi.Status.VolumeStatus = make([]v1.VolumeStatus, 1)
+
+		userData := "fake\nuser\ndata\n"
+		networkData := "FakeNetwork"
+		addCloudInitDisk(vmi, userData, networkData)
+		libvirtmanager.cloudInitDataStore = &cloudinit.CloudInitData{
+			DataSource: cloudinit.DataSourceNoCloud,
+			VolumeName: "test1",
+		}
+
+		err := libvirtmanager.generateCloudInitEmptyISO(vmi, nil)
+		Expect(err).To(HaveOccurred())
+		Expect(err.Error()).To(ContainSubstring("failed to find the status of volume test1"))
+	})
 
 	// TODO: test error reporting on non successful VirtualMachineInstance syncs and kill attempts
 
